#
#  AppController.rb
#  Sequence Clip Reporter
#
#  Created by Gregory Clarke on 7/23/09.
#  Copyright (c) 2009 Intelligent Assistance Software All rights reserved.
#
require 'osx/cocoa'
ENV['RUBYCOCOA_THREAD_HOOK_DISABLE'] = "true"
require 'yaml'
bundle = "#{File.dirname(__FILE__)}/sequence_to_spreadsheet.bundle"
if File.exist?(bundle)
	require 'sequence_to_spreadsheet.bundle'
else
	p "NO BUNDLE #{bundle} !!!"
end

class String
	def to_ascii
		ary = self.unpack("c*")
		ary.each_with_index do |x, i|
			if x < 31 || x > 127 # ASCII printable characters (character code 32-127)
				ary[i] = nil
			end
		end
		ascii_str = ary.compact.pack("c*")
		return ascii_str
	end
end

class AppController < OSX::NSObject

	ib_outlets :mainWindow, :deactivateMenuItem
	ib_outlets :view, :chooseXMLButton, :spinner, :fileIcon, :fileName, :sequencesPopupMenu, :analyzeButton, :videoTracksTableView, :videoTracksAllButton, :videoTracksNoneButton, :audioTracksTableView, :audioTracksAllButton, :audioTracksNoneButton, :enabledClipsRadiobuttons, :layoutsAppPopupMenu, :videoClipAudioCheckbox, :markersSheetsCheckbox, :generatorsSheetCheckbox, :filtersSheetCheckbox, :motionSheetCheckbox, :transitionsSheetCheckbox, :summarySheetCheckbox, :reelSheetsCheckbox, :openSpreadsheetCheckbox, :saveSpreadsheetButton
	ib_outlets :progressWindow, :progressView, :progressIndicator, :progressMessage
	ib_outlets :preferencesWindow, :preferencesTabs, :layoutsPrefPopupMenu, :addLayoutButton, :editLayoutButton, :removeLayoutButton, :rowNumbersColumnCheckbox, :thumbnailColumnCheckbox, :thumbnailLabel, :thumbnailWidthPopup, :keepThumbnailsCheckbox, :trackColumnCheckbox, :enabledColumnCheckbox, :inColumnCheckbox, :outColumnCheckbox, :durationColumnCheckbox, :srcReelColumnCheckbox, :srcInColumnCheckbox, :srcOutColumnCheckbox, :srcDurationColumnCheckbox, :soundRollColumnCheckbox, :descriptionColumnCheckbox, :sceneColumnCheckbox, :shottakeColumnCheckbox, :lognoteColumnCheckbox, :camerarollColumnCheckbox, :dailyrollColumnCheckbox, :labrollColumnCheckbox, :labelsColumnsCheckbox, :commentsColumnsCheckbox, :framerateColumnCheckbox, :formatColumnCheckbox, :srcFileNameColumnCheckbox, :srcFilePathColumnCheckbox, :metadataColumnCheckbox
	ib_outlets :layoutNameWindow, :layoutNameField, :layoutOKButton
	ib_outlets :importVideoAudioRadioButtons, :videoRowColorPopup, :audioRowColorPopup, :markerRowColorPopup
	ib_outlets :hhmmssffCheckbox1, :hhmmssffAdjField, :framesCheckbox1, :framesAdjField, :feetframesCheckbox1, :feetframesAdjField, :clockCheckbox1, :clockAdjField, :secondsCheckbox1, :secondsAdjField, :hhmmssffCheckbox2, :framesCheckbox2, :feetframesCheckbox2, :clockCheckbox2, :secondsCheckbox2
	ib_outlets :regoView, :webView, :webSpinner, :problemMessage, :closeRegoButton

	def initialize
		default_values = {"ReportHHMMSSFF1" => 1, "ReportHHMMSSFF2" => 1, "VideoRowsColor" => "black", "AudioRowsColor" => "black", "MarkerRowsColor" => "black", "AllOrEnabledClips" => 1, "Layout" => "Standard"}
		defaultsController = OSX::NSUserDefaultsController.sharedUserDefaultsController
		defaultsController.setInitialValues(default_values)
		nsAppKitVersionNumber10_6 = 1038
		nsAppKitVersionNumber10_9 = 1265
		nsAppKitVersionNumber10_10 = 1343
		nsAppKitVersionNumber10_14 = 1641 # https://wiki.freepascal.org/Cocoa_Internals/OS_Versions#AppKit_Version
		if OSX::NSAppKitVersionNumber.floor < nsAppKitVersionNumber10_10
			@is_yosemite_or_later = false
		else
			@is_yosemite_or_later = true
		end
		if OSX::NSAppKitVersionNumber.floor < nsAppKitVersionNumber10_14
			@is_mojave_or_later = false
			OSX.load_bridge_support_file "#{File.dirname(__FILE__)}/third-party/macOS.bridgesupport"
		else
			@is_mojave_or_later = true
		end
	end
		
	def awakeFromNib
		#puts "awakeFromNib"
		@spinner.stopAnimation(self) # otherwise it shows, even when set to not display when stopped
		@fileIcon.setImage(nil)
		@fileIcon.setToolTip("")
		@fileName.setStringValue("")
		@webView.setDrawsBackground(false)
		resources = OSX::NSBundle.mainBundle.resourcePath.fileSystemRepresentation
		@xml_icon = OSX::NSImage.alloc.initByReferencingFile(resources + "/XML.png")
		@ppro_icon = OSX::NSImage.alloc.initByReferencingFile(resources + "/PPro.png")
		@fcp_icon = OSX::NSImage.alloc.initByReferencingFile(resources + "/FCP.png")
		@fcp_version = "0"
@columnCheckboxes = [@rowNumbersColumnCheckbox, @thumbnailColumnCheckbox, @keepThumbnailsCheckbox, @trackColumnCheckbox, @enabledColumnCheckbox, @inColumnCheckbox, @outColumnCheckbox, @durationColumnCheckbox, @srcReelColumnCheckbox, @srcInColumnCheckbox, @srcOutColumnCheckbox, @srcDurationColumnCheckbox, @soundRollColumnCheckbox, @descriptionColumnCheckbox, @sceneColumnCheckbox, @shottakeColumnCheckbox, @lognoteColumnCheckbox, @camerarollColumnCheckbox, @dailyrollColumnCheckbox, @labrollColumnCheckbox, @labelsColumnsCheckbox, @commentsColumnsCheckbox, @framerateColumnCheckbox, @formatColumnCheckbox, @srcFileNameColumnCheckbox, @srcFilePathColumnCheckbox, @metadataColumnCheckbox]
		@columnCheckboxes.each { |x| x.setEnabled(false) }
		@editLayoutButton.setEnabled(false)
		@removeLayoutButton.setEnabled(false)
		validateTimeDisplayPreferences(self)
		paths = OSX::NSFileManager.defaultManager.URLsForDirectory_inDomains(OSX::NSApplicationSupportDirectory, OSX::NSUserDomainMask)
		$layouts_yaml = ""
		@layouts_array = []
		$checkboxes_yaml = ""
		@video_tracks_table_checkboxes, @video_tracks_table_checkboxes_clone = [], []
		@audio_tracks_table_checkboxes, @audio_tracks_table_checkboxes_clone = [], []
		#puts paths.inspect
		if paths.length > 0
			$app_support_subfolder = paths[0].path.stringByExpandingTildeInPath.to_s + "/com.AssistedEditing.SequenceClipReporter"
			Dir.mkdir($app_support_subfolder) unless File.directory?($app_support_subfolder)
			$inpoint_thumbnail_jpegs_folder = $app_support_subfolder + "/Thumbnails"
			Dir.mkdir($inpoint_thumbnail_jpegs_folder) unless File.directory?($inpoint_thumbnail_jpegs_folder)
			$layouts_yaml = $app_support_subfolder + "/Layouts.yml"
			if File.exist?($layouts_yaml)
				layouts_array = YAML.load_file($layouts_yaml)
				puts "layouts_array:", layouts_array.inspect
				if layouts_array==false # unable to load YAML file
				else # carefully build @layouts_array
					layouts_array.each do |l|
						unless l.nil? || l[:name].nil? || l[:name].empty?
							@layouts_array << l
							@layoutsPrefPopupMenu.addItemWithTitle(l[:name])
							@layoutsAppPopupMenu.addItemWithTitle(l[:name])
						end
					end
					# the selected Layout menuItem needs to be manually set from NSUserDefaults
					layout_name = OSX::NSUserDefaults.standardUserDefaults.stringForKey("Layout")
					puts "layout_name = #{layout_name.inspect}"
					layout_hash = @layouts_array.find { |l| l[:name]==layout_name.to_s }
					puts "Layout_hash = #{layout_hash.inspect}"
					if layout_name.nil? || layout_hash.nil?
						@layoutsAppPopupMenu.selectItemWithTitle("Standard")
					else
						@layoutsAppPopupMenu.selectItemWithTitle(layout_name) # this will add an item from NSUserDefaults that's not in the @layoutsPrefPopupMenu (built from @layouts_array), so do the layout_hash = @layouts_array.find check first
					end
					chooseLayoutApp(self) # and choose @layoutsPrefPopupMenu to match
				end
			end
			$checkboxes_yaml = $app_support_subfolder + "/Checkboxes.yml"
			if File.exist?($checkboxes_yaml)
				checkboxes_array = YAML.load_file($checkboxes_yaml)
				if checkboxes_array==false # unable to load YAML file
				elsif checkboxes_array.empty? == false
					@video_tracks_table_checkboxes_clone = checkboxes_array.first
					@audio_tracks_table_checkboxes_clone = checkboxes_array.last
				end
			end
		end
		#puts "@layouts_array #{@layouts_array.inspect}"
	end

	### rego methods ###
	
	def applicationWillFinishLaunching(sender) # applicationWillFinishLaunching: is sent before application:openFile: (so it can deal with a file dropped onto the app icon)
		#puts "applicationWillFinishLaunching"
		startup()
		if @this_rego[0]["active"]
			showClipReporterView()
		elsif @this_rego[0]["mac_problem"]==true
			check_for_rego(@this_rego)
		else
			displayDemoSheet()
		end
		OSX::LetsMove.PFMoveToApplicationsFolderIfNecessary # check_for_expired_beta THEN check LetsMove PFMoveToApplicationsFolderIfNecessary
		@openPanel = OSX::NSOpenPanel.openPanel # retain as instance variables
		@savePanel = OSX::NSSavePanel.savePanel
	end
	
	def applicationDidFinishLaunching(sender)
		#puts "applicationDidFinishLaunching"
		center = OSX::NSNotificationCenter.defaultCenter
		center.addObserver_selector_name_object(self, :resignKey, OSX::NSWindowDidResignKeyNotification, @preferencesWindow)
	end
	
	def displayDemoSheet
		@deactivateMenuItem.setTitle("Activate...")
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText("#{SKU} has not been activated for this computer.")
		alert.setAlertStyle(OSX::NSInformationalAlertStyle)
		alert.setInformativeText("Would you like to activate #{SKU}, or try it in demo mode (which is limited to 20 clips)?")
		alert.addButtonWithTitle("Activate")
		alert.addButtonWithTitle("Buy")
		alert.addButtonWithTitle("Demo")
		alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, "didEndDemoSheet:returnCode:contextInfo:", nil)
	end
	
	def didEndDemoSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked==OSX::NSAlertSecondButtonReturn
			system('open', 'https://intelligentassistance.com/sequence-clip-reporter-cc.html')
		end
		if buttonClicked==OSX::NSAlertThirdButtonReturn
			showClipReporterView()
		else
			newFrame = nil #self.newFrameForNewContentView(@regoView)
			self.appResizeWindow_forView(@mainWindow, @regoView)
			showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
		end
	end
	
	def showDeactivation
		if @deactivateMenuItem.title=="Activate..."
			newFrame = nil #self.newFrameForNewContentView(@regoView)
			self.appResizeWindow_forView(@mainWindow, @regoView)
			showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
		else
			newFrame = nil #self.newFrameForNewContentView(@regoView) # 500 x 200
			self.appResizeWindow_forView(@mainWindow, @regoView)
			@mainWindow.setMinSize(OSX::NSSize.new(500, 200+22))
			@mainWindow.setMaxSize(OSX::NSSize.new(500, 200+22))
			@mainWindow.setContentView(@regoView)
			showActivationView(newFrame, WEB_APP + "destroy/" + @this_rego[0]["number"].to_s)
		end
		@closeRegoButton.setTitle("Cancel")
	end
	ib_action :showDeactivation # Sequence Clip Reporter > Activate or Deactivate
	
	def showActivationView(frame, url)
		#puts "showActivationView #{url}"
		# @mainWindow.setFrame_display frame, true
		@mainWindow.setContentView(@regoView)
		@mainWindow.setTitle("Activation")
		# notification
		center = OSX::NSNotificationCenter.defaultCenter
		center.addObserver_selector_name_object(self, :webProgress, nil, @webView)
		@webView.mainFrame.loadRequest( OSX::NSURLRequest.requestWithURL( OSX::NSURL.URLWithString(url) ) )
	end
	
	def webProgress(notification)
		case notification.name
		when "WebProgressStartedNotification"
			@webSpinner.startAnimation(self)
		when "WebProgressFinishedNotification"
			@webSpinner.stopAnimation(self)
			if @webView.mainFrame.dataSource.nil? # in case page doesn't load...
				@webView.mainFrame.stopLoading
				@problemMessage.setHidden(false)
				@webView.setHidden(true)
			else
				get_web_app_response(@webView.mainFrame.dataSource.response.URL)
			end
		end
	end 
	
	def closeActivationView
		if @closeRegoButton.title=="Continue" || @closeRegoButton.title=="Cancel"
			showClipReporterView()
		else
			OSX::NSApp.terminate(self)
		end
	end
	ib_action :closeActivationView # on click of Registration window's Continue button
	
	### end rego methods ###
	
	def showClipReporterView
		setupView(@view, "Sequence Clip Reporter", 400, 600)
	end
	
	def setupView(view, title, width, height)
		@mainWindow.setTitle(title)
#		newFrame = self.newFrameForNewContentView(view)
#		@mainWindow.setMinSize(OSX::NSSize.new(width, height+22))
#		@mainWindow.setFrame_display_animate(newFrame, true, true)
		self.appResizeWindow_forView(@mainWindow, view)
		@mainWindow.setContentView(view)
	end
	
#	def newFrameForNewContentView(view)
#		newFrameRect = @mainWindow.frameRectForContentRect(view.frame)
#		oldFrameRect = @mainWindow.frame
#		newSize = newFrameRect.size
#		oldSize = oldFrameRect.size
#		frame = @mainWindow.frame
#		frame.size = newSize
#		frame.origin.y = frame.origin.y - (newSize.height - oldSize.height)
#		frame
#	end

	def showHelp(sender)
		openPDF("Sequence Clip Reporter - Help.pdf")
	end
	ib_action :showHelp # Help > Sequence Clip Reporter Help
	
	def openPDF(filename)
		path_to_file = __FILE__.sub("AppController.rb",filename)
		system('open', path_to_file)
	end
	
	def checkForProVideoFormats(sender)
		alert = OSX::NSAlert.alloc.init
		if File.exist?("/Library/Video/Professional Video Workflow Plug-Ins/AppleMXFImport.bundle") 
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			alert.setMessageText("Success! Pro Video Formats are installed")
			alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
		else
			alert.setAlertStyle(OSX::NSCriticalAlertStyle)
			alert.setMessageText("Pro Video Formats are NOT installed")
			alert.addButtonWithTitle("OK")
			alert.addButtonWithTitle("Get Pro Video Formats")
			buttonClicked = alert.runModal
			if buttonClicked==1001
				system('open', 'https://support.apple.com/kb/index?page=search&q=%22pro%20video%20formats%22&currentPage=1&includeArchived=false&locale=en_US&src=globalnav_support&type=suggested')
			end
		end
	end
	ib_action :checkForProVideoFormats
	
	def showAboutThumbnails(sender)
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText("Thumbnail images can only be created from QuickTime/AV Foundation-compatible media that is online")
		alert.setAlertStyle(OSX::NSInformationalAlertStyle)
		alert.setInformativeText("Sequence Clip Reporter running on Mac OS X 10.10 or later supports more media types if the free “Pro Video Formats” are installed.")
		alert.addButtonWithTitle("OK")
		alert.addButtonWithTitle("Get Pro Video Formats")
		buttonClicked = alert.runModal
		if buttonClicked==1001
			system('open', 'https://support.apple.com/kb/index?page=search&q=%22pro%20video%20formats%22&currentPage=1&includeArchived=false&locale=en_US&src=globalnav_support&type=suggested')
		end
	end
	ib_action :showAboutThumbnails # click the Help button in Preferences
	
	def emailSupport(sender)
    executable = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleExecutable")
		version = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleShortVersionString")
		app = "#{executable} (version #{version})".to_ns.stringByAddingPercentEscapesUsingEncoding(4)
		body = "What went wrong?\n\n\nAttached is the XML file that caused the problem:\n\n\n".to_ns.stringByAddingPercentEscapesUsingEncoding(4)
		mailto = OSX::NSURL.URLWithString("mailto:info@IntelligentAssistance.com?subject=Bug%20Report%20for%20#{app}&body=#{body}")
		result = OSX::NSWorkspace.sharedWorkspace.openURL(mailto)
		if result==false
			contactus = OSX::NSURL.URLWithString("https://intelligentassistance.com/contact-us.html")
			OSX::NSWorkspace.sharedWorkspace.openURL(contactus)
		end
	end
	ib_action :emailSupport

	def visitWebsite(sender)
		system('open', 'https://intelligentassistance.com/sequence-clip-reporter-cc.html')
	end
	ib_action :visitWebsite # Help > Visit Web Site

	def sendSupportEmail(subject)
		system('open', "mailto:info@IntelligentAssistance.com?subject=Sequence Clip Reporter #{subject}&body=Attached is the XML file that Sequence Clip Reporter was unable to process.")
	end

	# def applicationDidBecomeActive(notification) # Snow Leopard
	# 	#puts "applicationDidBecomeActive"
	# 	if @mainWindow.contentView==@view && @nokogiri_xml.nil?
	# 		# @spinner.startAnimation(self)
	# 		proj = get_project_name()
	# 		#puts "@project_filepath = #{@project_filepath}", proj.inspect
	# 		if proj=="Final Cut Pro is not installed" || proj.empty?
	# 			@chooseXMLButton.setEnabled(true)
	# 		else
	# 			@chooseXMLButton.setEnabled(false)
	# 			@fileIcon.setImage(@fcp_icon)
	# 			@fileIcon.setToolTip(proj)
	# 			@fileName.setStringValue(proj)
	# 		end
	# 		# @spinner.stopAnimation(self)
	# 	end
	# end

	def chooseXmlFile(sender)
		filetypes = ["xml"]
		@openPanel.setMessage("Choose the exported XML file")
		@openPanel.setAllowsMultipleSelection(false)
		@openPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseXmlFile

	def didEndOpenSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		unless @openPanel.URL.nil?
			# @openPanel.orderOut(self)
			path = @openPanel.URL.path.to_s
			application_openFile(self, path)
		end
	end

	def application_openFile(sender, path)
		chosen_file = path.to_s
		chosen_file_extension = File.extname(chosen_file)
		@fileIcon.setToolTip(chosen_file)
		@fileName.setStringValue(File.basename(chosen_file))
		@fileIcon.setNeedsDisplay(true)
		@fileName.setNeedsDisplay(true)
		@spinner.startAnimation(self)
		# Thread.new { check_the_mac(@this_rego[0]["number"].to_s) }
		begin
			@nokogiri_xml, doctype, @is_ppro = parse_xml(File.open(chosen_file))
			if doctype!= "xmeml"
				raise "not xmeml"
			end
			#puts @is_ppro
			if @is_ppro
				@fileIcon.setImage(@ppro_icon)
			else
				@fileIcon.setImage(@xml_icon)
			end
			if @nokogiri_xml.xpath("//sequence/uuid").empty? # XML with no sequences or uuids
				@nokogiri_xml.xpath("//sequence").each_with_index do |seq, i|
					if seq.at_xpath("uuid").nil?
						seq.add_child("<uuid>UUID#{i}</uuid>")
					end
				end
			end
			get_sequences_names(@nokogiri_xml.to_s)
			@spinner.stopAnimation(self)
			OSX::NSDocumentController.sharedDocumentController().noteNewRecentDocumentURL( OSX::NSURL.fileURLWithPath_isDirectory(path, false) ) # add this file to the Recent menu
			return true
		rescue => e
			if $ruby_cocoa==false
				raise e
			else
				@spinner.stopAnimation(self)
	      executable = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleExecutable")
				version = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleShortVersionString")
				alert = OSX::NSAlert.alloc.init
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				case e.message
				when "fatal error"
					alert.setMessageText("This XML file is incomplete or broken")
					alert.setInformativeText("Please export the XML file again.")
				when "not xmeml"
					alert.setMessageText("This is not a Final Cut Pro XML file")
					alert.setInformativeText("Please choose an XML file exported from\nPremiere Pro CC or Final Cut Pro 7.")
				else
					OSX::NSLog "#{e.message}\n#{e.backtrace.join("\n")}"
					alert.setMessageText("There was a problem converting this XML#{@error_tracker}")
					alert.setInformativeText("Please zip up the XML file and email it with the subject “#{executable} #{version} error” to info@assistedediting.com with this error message in the body text:\n
#{e.message}\n
Thanks for helping us improve #{executable}!")
				end
				alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
			end
		end
	end
	
	def switchSequence(sender)
		#puts "switchSequence"
		unless @sequence_uuids.length==1
			if @sequencesPopupMenu.indexOfSelectedItem==@analyzed_sequence_index # reselected an analyzed sequence
				@video_tracks_table_checkboxes = @video_tracks_table_checkboxes_clone
				@audio_tracks_table_checkboxes = @audio_tracks_table_checkboxes_clone
				enableSaveSpreadsheetButton()
			else
				clone_and_clear_tableview_checkboxes()
				@saveSpreadsheetButton.setEnabled(false)
			end
			@videoTracksTableView.reloadData
			@videoTracksTableView.deselectRow(0)
			@audioTracksTableView.reloadData
			@audioTracksTableView.deselectRow(0)
		end
	end
	ib_action :switchSequence
	
	def analyzeSequence(sender)
		clone_and_clear_tableview_checkboxes()
		@videoTracksTableView.reloadData
		@audioTracksTableView.reloadData
		Thread.new { analyze_sequence() }
	end
	ib_action :analyzeSequence
	
	def analyze_sequence
		puts "analyzeSequence"
		@analyzed_sequence_index = @sequencesPopupMenu.indexOfSelectedItem
		puts "@analyzed_sequence_index = #{@analyzed_sequence_index}"
		begin
			displayProgressSheet() # changed 10/4/2021
			@progressIndicator.setIndeterminate(true)
			# @progressIndicator.setUsesThreadedAnimation(true)
			@progressIndicator.startAnimation(self)
			@progressIndicator.displayIfNeeded
			@progressMessage.setStringValue("Opening XML file...\n")
			@progressMessage.displayIfNeeded
			# get uuid to analyze
			uuid = @sequence_uuids[@sequencesPopupMenu.indexOfSelectedItem]
			# puts "  uuid = #{uuid.inspect}"
			puts "get_info_from_sequence_or_bins(#{uuid})"
			get_info_from_sequence_or_bins(uuid)
			puts @video_tracks_table_checkboxes.inspect, @audio_tracks_table_checkboxes.inspect
			# load tracks into tables
			@videoTracksTableView.reloadData
			if @video_tracks_table_checkboxes.length > 0
				@videoTracksTableView.deselectRow(0)
				@videoTracksAllButton.setEnabled(true)
				@videoTracksNoneButton.setEnabled(true)
			else
				@videoTracksAllButton.setEnabled(false)
				@videoTracksNoneButton.setEnabled(false)
			end
			@audioTracksTableView.reloadData
			if @audio_tracks_table_checkboxes.length > 0
				@audioTracksTableView.deselectRow(0)
				@audioTracksAllButton.setEnabled(true)
				@audioTracksNoneButton.setEnabled(true)
			else
				@audioTracksAllButton.setEnabled(false)
				@audioTracksNoneButton.setEnabled(false)
			end
			enableSaveSpreadsheetButton()
		rescue => e
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			case e.message
			when "Memory allocation failed : growing nodeset hit limit: growing nodeset hit limit\n"
				alert.setMessageText("This XML file is too big")
				alert.setInformativeText("Try selecting a single sequence before exporting XML.")
			else
				OSX::NSLog "#{e.message}\n#{e.backtrace.join("\n")}"
				executable = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleExecutable")
				version = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleShortVersionString")
				alert.setMessageText("There was a problem reading the XML file of this sequence.")
				alert.setInformativeText("Please zip up the XML file and email it with the subject “#{executable} #{version} error” to info@assistedediting.com with this error message in the body text:\n
#{e.message}\n
Thanks for helping us improve #{executable}!")
			end
			didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
			alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
		end
	end
	
	def displayProgressSheet
		@spinner.stopAnimation(self)
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo(@progressWindow, @mainWindow, self, nil, nil)
		@progressView.displayIfNeeded
	end

	def didEndProgressSheet_returnCode_contextInfo(sheet, code, info)
		puts "didEndProgressSheet_returnCode_contextInfo code #{code}"
		@progressMessage.setStringValue("")
		@progressIndicator.stopAnimation(nil)
		sheet.orderOut(self)
		@mainWindow.endSheet(sheet) # OSX::NSApp.endSheet(sheet)
		OSX::NSApp.dockTile.setBadgeLabel("")
		# if code < 0 # error condition
		# 	@saveMenuItem.setEnabled(false)
		# else
		# 	@saveMenuItem.setEnabled(true)
		# end
	end

	def selectAllVideoTracks(sender)
		@video_tracks_table_checkboxes.collect! { |x| x = 1 }
		@videoTracksTableView.reloadData
		@videoTracksTableView.deselectRow(0)
		enableSaveSpreadsheetButton()
	end
	ib_action :selectAllVideoTracks

	def selectNoVideoTracks(sender)
		@video_tracks_table_checkboxes.collect! { |x| x = 0 }
		@videoTracksTableView.reloadData
		@videoTracksTableView.deselectRow(0)
		enableSaveSpreadsheetButton()
	end
	ib_action :selectNoVideoTracks

	def selectAllAudioTracks(sender)
		@audio_tracks_table_checkboxes.collect! { |x| x = 1 }
		@audioTracksTableView.reloadData
		@audioTracksTableView.deselectRow(0)
		enableSaveSpreadsheetButton()
	end
	ib_action :selectAllAudioTracks

	def selectNoAudioTracks(sender)
		@audio_tracks_table_checkboxes.collect! { |x| x = 0 }
		@audioTracksTableView.reloadData
		@audioTracksTableView.deselectRow(0)
		enableSaveSpreadsheetButton()
	end
	ib_action :selectNoAudioTracks
	
	# control whether the Save Spreadsheet button is enabled or not
	def enableSaveSpreadsheetButton
		if @video_tracks_table_checkboxes.include?(1) || @audio_tracks_table_checkboxes.include?(1) # there's a checkbox selected
			if @hhmmssffCheckbox1.state==0 && @framesCheckbox1.state==0 && @feetframesCheckbox1.state==0 && @secondsCheckbox1.state==0 && @clockCheckbox1.state==0
				@saveSpreadsheetButton.setEnabled(false)
			elsif @hhmmssffCheckbox2.state==0 && @framesCheckbox2.state==0 && @feetframesCheckbox2.state==0 && @secondsCheckbox2.state==0 && @clockCheckbox2.state==0
				@saveSpreadsheetButton.setEnabled(false)
			else
				@saveSpreadsheetButton.setEnabled(true)
			end
		else
			@saveSpreadsheetButton.setEnabled(false)
		end
	end
	ib_action :enableSaveSpreadsheetButton

	def saveSpreadsheet(sender)
		@savePanel.setTitle("Save the Spreadsheet")
		@savePanel.setRequiredFileType("xls")
		@savePanel.setExtensionHidden(false)
		@savePanel.setCanSelectHiddenExtension(true)
		item = @sequencesPopupMenu.titleOfSelectedItem
		layout = @layoutsAppPopupMenu.titleOfSelectedItem
		@savePanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{item} - #{layout} Report", @mainWindow, self, 'didEndSaveSpreadsheet:returnCode:contextInfo:', nil)
		saveYAMLs()
	end
	ib_action :saveSpreadsheet

	def didEndSaveSpreadsheet_returnCode_contextInfo(sheet, buttonClicked, info)
		if buttonClicked==1 # NSOKButton
			@savePanel.orderOut(self)
			if @savePanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			excel_file = @savePanel.URL.path.to_s
			@report_enabled_only = case @enabledClipsRadiobuttons.selectedCell.tag
			when 0 then true
			when 1 then false
			end
			@timeline_time_columns = []
			@timeline_time_columns << "Timecode" if @hhmmssffCheckbox1.state==1
			@timeline_time_columns << "Frames" if @framesCheckbox1.state==1
			@timeline_time_columns << "Feet+Frames" if @feetframesCheckbox1.state==1
			@timeline_time_columns << "Clock" if @clockCheckbox1.state==1
			@timeline_time_columns << "Seconds" if @secondsCheckbox1.state==1
			@hhmmssff_adj = @hhmmssffAdjField.doubleValue.to_i
			@frames_adj = @framesAdjField.doubleValue.to_i
			@feetframes_adj = @feetframesAdjField.doubleValue.to_i
			@clock_adj = @clockAdjField.doubleValue.to_i
			@seconds_adj = @secondsAdjField.doubleValue.to_i
			@source_time_columns = []
			@source_time_columns << "Timecode" if @hhmmssffCheckbox2.state==1
			@source_time_columns << "Frames" if @framesCheckbox2.state==1
			@source_time_columns << "Feet+Frames" if @feetframesCheckbox2.state==1
			@source_time_columns << "Clock" if @clockCheckbox2.state==1
			@source_time_columns << "Seconds" if @secondsCheckbox2.state==1
			@inpoint_thumbnail_jpegs = case @thumbnailColumnCheckbox.state
			when 0 then false
			when 1 then true
			end
			@thumbnail_width = @thumbnailWidthPopup.stringValue.to_i # .titleOfSelectedItem.to_i
			save_excel_report(excel_file)
		end
	end

	def save_excel_report(excel_file)
		displayProgressSheet() # changed 10/4/2021
		@progressIndicator.setIndeterminate(true)
		# @progressIndicator.setUsesThreadedAnimation(true)
		@progressIndicator.startAnimation(self)
		@progressIndicator.displayIfNeeded
		@progressMessage.setStringValue("Creating report...\n")
		@progressMessage.displayIfNeeded
		GC.disable # temporarily disable garbage collection to prevent GC-related errors when executing from a compiled bundle, esp. NSImage
		########## MAKE AND SAVE EXCEL ##########
		begin
			save_excel(excel_file)
			if @openSpreadsheetCheckbox.state==1
				OSX::NSWorkspace.sharedWorkspace.openFile(excel_file)
			end
		rescue
			didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
			executable = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleExecutable")
			version = OSX::NSBundle.mainBundle.infoDictionary.objectForKey("CFBundleShortVersionString")
			alert = OSX::NSAlert.alloc.init
			alert.setMessageText("There was a problem saving a Spreadsheet of these clips.")
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			#puts "!!! Problem with saveExcel()"
			if @current_thumbnail_clip.nil?
				alert.setInformativeText("Please zip up the .xml file and email it with the subject \"#{executable} #{version} error\" to info@assistedediting.com - thanks for helping us improve #{executable}!")
			else
				alert.setInformativeText("Creating a thumbnail image for the clip #{@current_thumbnail_clip} raised this error.")
			end
			alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, nil, nil) # Deprecated after 10.10
			# sendSupportEmail("problem saving Excel spreadsheet") if buttonClicked==OSX::NSAlertFirstButtonReturn
		end
		GC.enable # re-enable garbage collection
		GC.start # and initiate garbage collection
	end

	### TableView methods ###

	def numberOfRowsInTableView(tableView)
		if tableView.tag==1 # video tracks
			if @video_tracks_table_checkboxes.nil?
				return 0
			else
				return @video_tracks_table_checkboxes.length
			end
		elsif tableView.tag==2 # audio tracks
			if @audio_tracks_table_checkboxes.nil?
				return 0
			else
				return @audio_tracks_table_checkboxes.length
			end
		end
	end

	def tableView_objectValueForTableColumn_row(tableView, tableColumn, rowIndex)
		if tableView.tag==1 # video tracks
			tableColumn.dataCellForRow(rowIndex).setTitle("Track V#{rowIndex+1}")
			return @video_tracks_table_checkboxes[rowIndex]
		elsif tableView.tag==2 # audio tracks
			tableColumn.dataCellForRow(rowIndex).setTitle("Track A#{rowIndex+1}")
			return @audio_tracks_table_checkboxes[rowIndex]
		end
	end

	def tableView_setObjectValue_forTableColumn_row(tableView, anObject, tableColumn, rowIndex)
		if tableView.tag==1 # video tracks
			tableView.deselectRow(rowIndex)
			@video_tracks_table_checkboxes[rowIndex] = anObject.to_i
		elsif tableView.tag==2 # audio tracks
			tableView.deselectRow(rowIndex)
			@audio_tracks_table_checkboxes[rowIndex] = anObject.to_i
		end
		enableSaveSpreadsheetButton()
	end

	### Layouts methods ###
	
	def resignKey(notification) # Preferences window has been closed
		updateLayoutsArray(@thumbnailWidthPopup) # in case the user didn't end editing before closing the Preferences window
	end
	
	def showLayoutsPref(sender)
		@preferencesWindow.makeKeyAndOrderFront(self)
		@preferencesTabs.selectFirstTabViewItem(self)
	end
	ib_action :showLayoutsPref
	
	def chooseLayoutApp(sender) # change popup menu in App
		@layoutsPrefPopupMenu.selectItemAtIndex(@layoutsAppPopupMenu.indexOfSelectedItem) # keep the popup menus in sync
		chooseLayoutPref(self)
	end
	ib_action :chooseLayoutApp
	
	def chooseLayoutPref(sender) # change popup menu in Prefs
		# puts "chooseLayoutPref"
		@layoutsAppPopupMenu.selectItemAtIndex(@layoutsPrefPopupMenu.indexOfSelectedItem) # keep the popup menus in sync
		@thumbnailWidthPopup.selectItemWithObjectValue("40")
		if @layoutsPrefPopupMenu.indexOfSelectedItem==0 # all on except row numbers, thumbnails, sound roll, ID3 metadata and comments
			@layoutsAppPopupMenu.selectItemAtIndex(0)
			@columnCheckboxes.each { |x| x.setEnabled(false); x.setState(1) }
			@rowNumbersColumnCheckbox.setState(0) unless @rowNumbersColumnCheckbox.nil?
			@thumbnailColumnCheckbox.setState(0) unless @thumbnailColumnCheckbox.nil?
			@keepThumbnailsCheckbox.setState(0) unless @keepThumbnailsCheckbox.nil?
			@soundRollColumnCheckbox.setState(0) unless @soundRollColumnCheckbox.nil?
			@metadataColumnCheckbox.setState(0) unless @metadataColumnCheckbox.nil?
			@commentsColumnsCheckbox.setState(0) unless @commentsColumnsCheckbox.nil?
			@thumbnailWidthPopup.setEnabled(false)
			@keepThumbnailsCheckbox.setEnabled(false)
			@editLayoutButton.setEnabled(false)
			@removeLayoutButton.setEnabled(false)
		else
			@columnCheckboxes.each { |x| x.setEnabled(true) }
			# if @is_mavericks==false # might work for Lion or later?
			# 	@thumbnailColumnCheckbox.setState(0)
			# 	@thumbnailColumnCheckbox.setEnabled(false)
			# 	@thumbnailLabel.setEnabled(false)
			# 	@thumbnailWidthPopup.setEnabled(false)
			# end
			@thumbnailWidthPopup.setEnabled(true)
			@keepThumbnailsCheckbox.setEnabled(true)
			@editLayoutButton.setEnabled(true)
			@removeLayoutButton.setEnabled(true)
			layout_hash = @layouts_array.find { |l| l[:name]==@layoutsPrefPopupMenu.titleOfSelectedItem.to_s }
			# pp layout_hash
			if layout_hash.nil?
				puts "Layout_hash is NIL"
			else
				layout_items = @columnCheckboxes + [@thumbnailWidthPopup]
				layout_items.each do |x|
					if x.tag==1
						title = "thumbnailWidthPopup"
					else
						title = x.title.to_s
					end
					checkbox_hash = layout_hash[:checkboxes].find { |c| c[:title]==title }
					# pp checkbox_hash
					if title == "thumbnailWidthPopup"
						if checkbox_hash != nil && @thumbnailWidthPopup != nil
							puts "@thumbnailWidthPopup.setStringValue(#{checkbox_hash[:state].inspect})"
							@thumbnailWidthPopup.setStringValue(checkbox_hash[:state])
							@thumbnailWidthPopup.selectItemWithObjectValue(checkbox_hash[:state])
						end
					elsif checkbox_hash.nil? || checkbox_hash[:state]==0 # missing checkboxes have not been clicked on
						x.setState(0)
					else checkbox_hash[:state]==1
						x.setState(1)
					end
				end
			end
		end
	end
	ib_action :chooseLayoutPref
	
	def addLayout(sender)
		# puts "addLayout"
		@layoutNameField.setStringValue("")
		@layoutOKButton.setTag(1)
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo(@layoutNameWindow, @preferencesWindow, self, nil, nil)
	end
	ib_action :addLayout
	
	def editLayout(sender)
		layout_name = @layoutsPrefPopupMenu.titleOfSelectedItem.to_s
		@layoutNameField.setStringValue(layout_name)
		@layoutOKButton.setTag(2)
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo(@layoutNameWindow, @preferencesWindow, self, nil, nil)
	end
	ib_action :editLayout
	
	def updateLayoutsPopup(sender) # sent by OK button in Name the Layout window
		# puts "updateLayoutsPopup"
		@layoutNameWindow.orderOut(self)
		@preferencesWindow.endSheet(@layoutNameWindow) # OSX::NSApp.endSheet(@layoutNameWindow)
		old_layout_name = @layoutsPrefPopupMenu.titleOfSelectedItem.to_s
		new_layout_name = @layoutNameField.stringValue.to_s
		unless new_layout_name.empty?
			if old_layout_name==new_layout_name
				snap = nil # name wasn't edited
			else
				snap = @layouts_array.find { |l| l[:name]==new_layout_name }
			end
			# can't have 2 popup menu items with the same name
			unless snap.nil?
				new_layout_name = "#{new_layout_name} 1"
				until snap.nil?
					snap = @layouts_array.find { |l| l[:name]==new_layout_name }
					unless snap.nil?
						new_layout_name.succ!
					end
				end
			end
			if sender.tag==1 # add button
				# puts "add"
				@layoutsPrefPopupMenu.addItemWithTitle(new_layout_name)
				@layoutsPrefPopupMenu.selectItemWithTitle(new_layout_name)
				@layoutsAppPopupMenu.addItemWithTitle(new_layout_name)
				@columnCheckboxes.each do |x|
					x.setEnabled(true)
					x.setState(0)
				end
				@thumbnailWidthPopup.setEnabled(true) # will be disabled if Standard was previously selected
				@keepThumbnailsCheckbox.setEnabled(true)
				# if @is_mavericks==false
				# 	@thumbnailColumnCheckbox.setEnabled(false)
				# 	@thumbnailLabel.setEnabled(false)
				# 	@thumbnailWidthPopup.setEnabled(false)
				# end
				@editLayoutButton.setEnabled(true)
				@removeLayoutButton.setEnabled(true)
				@layouts_array << { :name => new_layout_name, :checkboxes => [] }
			elsif old_layout_name!=new_layout_name # edit button, and they're different names
				# puts "edit, new name"
				layout_hash = @layouts_array.find { |l| l[:name]==old_layout_name }
				layout_hash[:name] = new_layout_name
				index = @layoutsPrefPopupMenu.indexOfSelectedItem
				@layoutsPrefPopupMenu.insertItemWithTitle_atIndex(new_layout_name, index)
				@layoutsPrefPopupMenu.selectItemWithTitle(new_layout_name)
				@layoutsPrefPopupMenu.removeItemWithTitle(old_layout_name)
				if @layoutsAppPopupMenu.indexOfSelectedItem==index
					reselect = true
				else
					reselect = false
				end
				@layoutsAppPopupMenu.insertItemWithTitle_atIndex(new_layout_name, index)
				@layoutsAppPopupMenu.selectItemWithTitle(new_layout_name) if reselect
				@layoutsAppPopupMenu.removeItemWithTitle(old_layout_name)
			else
				# puts "edit, same names"
			end
		end
		saveYAMLs()
	end
	ib_action :updateLayoutsPopup
	
	def removeLayout(sender)
		layout_name = @layoutsPrefPopupMenu.titleOfSelectedItem.to_s
		layout_hash = @layouts_array.find { |l| l[:name]==layout_name }
		@layoutsPrefPopupMenu.removeItemWithTitle(layout_name)
		@layoutsAppPopupMenu.removeItemWithTitle(layout_name)
		@layouts_array.delete(layout_hash)
		# puts @layouts_array.inspect
		# first item becomes selected
		chooseLayoutPref(self)
		saveYAMLs()
	end
	ib_action :removeLayout
	
	def updateLayoutsArray(sender) # click a layout checkbox in Prefs
		if @layoutsPrefPopupMenu.indexOfSelectedItem > 0 # no need to update @layouts_array is Standard is selected
			layout_hash = @layouts_array.find { |l| l[:name]==@layoutsPrefPopupMenu.titleOfSelectedItem.to_s }
			if sender.tag==1 # @thumbnailWidthPopup
				title = "thumbnailWidthPopup"
				int = @thumbnailWidthPopup.stringValue.to_i
				if int==0 # a non-numberic.to_i will be nil, and nil.to_s will be 0
					state = "40"
				else
					state = int.abs.to_s
				end
			else
				title = sender.title.to_s
				state = sender.state.to_i
			end
			checkbox_hash = layout_hash[:checkboxes].find { |c| c[:title]==title }
			layout_hash[:checkboxes].delete(checkbox_hash)
			layout_hash[:checkboxes] << { :title => title, :state => state }
			saveYAMLs()
		end
	end
	ib_action :updateLayoutsArray
	
	def saveYAMLs
		File.open($layouts_yaml, "w") { |f| YAML.dump(@layouts_array, f) }
		if @video_tracks_table_checkboxes!=nil && @audio_tracks_table_checkboxes!=nil
			checkboxes_array = [@video_tracks_table_checkboxes, @audio_tracks_table_checkboxes]
			File.open($checkboxes_yaml, "w") { |f| YAML.dump(checkboxes_array, f) }
		end
	end
	
	# ensure that at least one time checkbox is on
	def validateTimeDisplayPreferences(sender)
		if @hhmmssffCheckbox1.state==0 && @framesCheckbox1.state==0 && @feetframesCheckbox1.state==0 && @secondsCheckbox1.state==0 && @clockCheckbox1.state==0
			@hhmmssffCheckbox1.state = 1
		end
		if @hhmmssffCheckbox2.state==0 && @framesCheckbox2.state==0 && @feetframesCheckbox2.state==0 && @secondsCheckbox2.state==0 && @clockCheckbox2.state==0
			@hhmmssffCheckbox2.state = 1
		end
	end
	ib_action :validateTimeDisplayPreferences
	
	# def applicationShouldTerminate(sender)
	# => Don't saveYAMLs here because a crash can corrupt the file
	# 	return true
	# end
	
	def applicationShouldTerminateAfterLastWindowClosed(sender)
		true
	end

end
